#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include "socket.h"

static void error_exit(char *message)
{
  fprintf(stderr, message);
  exit(1);
}

int socket_open_server(int port)
{
  int s, sockt;
  struct sockaddr_in addr;
  socklen_t addrlen;
  int on = 1;

  s = socket(AF_INET, SOCK_STREAM, 0);
  if (s < 0)
    error_exit("socket() failed.\n");

  if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0)
    error_exit("setsockopt() failed.\n");

  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  addr.sin_addr.s_addr = htonl(INADDR_ANY);

  if (bind(s, (struct sockaddr *)&addr, sizeof(addr)) < 0)
    error_exit("bind() failed.\n");

  if (listen(s, 5) < 0)
    error_exit("listen() failed.\n");

  addrlen = sizeof(addr);
  sockt = accept(s, (struct sockaddr *)&addr, &addrlen);
  if (sockt < 0)
    error_exit("accept() failed.\n");

  close(s);

  return sockt;
}

int socket_open_client(char *hostname, int port)
{
  int s;
  struct hostent *host;
  struct sockaddr_in addr;

  host = gethostbyname(hostname);
  if (host == NULL)
    error_exit("gethostbyname() failed.\n");

  s = socket(AF_INET, SOCK_STREAM, 0);
  if (s < 0)
    error_exit("socket() failed.\n");

  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  memcpy(&addr.sin_addr, host->h_addr, host->h_length);

  if (connect(s, (struct sockaddr *)&addr, sizeof(addr)) < 0)
    error_exit("connect() failed.\n");

  return s;
}

int socket_read(int s, char *buffer, int size, int wait_usec)
{
  fd_set fds;
  struct timeval t;
  int r;

  if (wait_usec >= 0) {
    FD_ZERO(&fds);
    FD_SET(s, &fds);
    t.tv_sec  = wait_usec / 1000000;
    t.tv_usec = wait_usec % 1000000;
    r = select(s + 1, &fds, NULL, NULL, &t);
    if (r <= 0)
      return r;
    if (!FD_ISSET(s, &fds))
      return 0;
  }

  return read(s, buffer, size);
}

int socket_write(int s, char *buffer, int size)
{
  return write(s, buffer, size);
}

static char *skip_space(char *p)
{
  while (*p) {
    if (!isspace(*p))
      break;
    p++;
  }
  return p;
}

static char *skip_nospace(char *p)
{
  while (*p) {
    if (isspace(*p))
      break;
    p++;
  }
  return p;
}

char **socket_readline(int s, int wait_usec)
{
  static int buffer_offset = 0;
  static int buffer_size = 0;
  static char buffer[1024];
  static char *argv[16];
  int r, i;
  char *p;

  if (buffer_offset > 0) {
    for (i = 0; i < buffer_size; i++) {
      buffer[i] = buffer[buffer_offset + i];
    }
    buffer_offset = 0;
  }

  r = socket_read(s, buffer + buffer_size,
		  sizeof(buffer) - buffer_size, wait_usec);
  if (r < 0)
    return NULL;

  buffer_size += r;

  for (i = 0; i < buffer_size; i++) {
    if ((buffer[i] == '\n') || (buffer[i] == '\r'))
      break;
  }
  if (i == buffer_size) {
    if (buffer_size == sizeof(buffer))
      buffer_size = 0;
    return NULL;
  }
  buffer[i] = '\0';

  buffer_offset = i + 1;
  buffer_size -= buffer_offset;

  p = buffer;
  i = 0;

  while (1) {
    p = skip_space(p);
    if (*p == '\0')
      break;

    argv[i++] = p;
    if (i == sizeof(argv) / sizeof(argv[0]))
      return NULL;

    p = skip_nospace(p);
    if (*p == '\0')
      break;

    *(p++) = '\0';
  }

  argv[i] = NULL;

  return argv;
}
